package net.sf.alchim.codeplus.spoonprocessor;

import net.sf.alchim.codeplus.annotation.Default;
import net.sf.alchim.codeplus.annotation.Nullable;
import net.sf.alchim.codeplus.annotation.ParamsNotNull;

import spoon.processing.AbstractProcessor;
import spoon.reflect.code.CtBlock;
import spoon.reflect.code.CtExpression;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtConstructor;
import spoon.reflect.declaration.CtExecutable;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.declaration.CtParameter;
import spoon.template.Substitution;

public class NullParamProcessor extends AbstractProcessor<CtExecutable<?>> {

    public void process(CtExecutable<?> m) {
        if (!(m instanceof CtMethod)
                && !(m instanceof CtConstructor)) {
            return;
        }
        addDefaulting(m);
        addNullChecker(m);
        removeAnnotations(m);
    }

    // TODO: add defaulting to same type as param
    protected void addDefaulting(CtExecutable<?> m) {
        for (CtParameter<?> param : m.getParameters()) {
            Default annotation = param.getAnnotation(Default.class);
            if (annotation == null) {
                continue;
            }
            if ((annotation.value() == null)
                    || (annotation.value().length() == 0)) {
                continue;
            }
            CtExpression<Object> defaultExpr = getFactory().Code().createCodeSnippetExpression(annotation.value());
            CtBlock<?> defaultBlock = Substitution.substituteMethodBody((CtClass<?>) m.getDeclaringType(), new NullParamTemplate(param, defaultExpr), "setDefault", getFactory().Type()
                    .createReference(Object.class));
            m.getBody().insertBegin(defaultBlock);
        }
    }

    protected void addNullChecker(CtExecutable<?> m) {
        ParamsNotNull annotation = m.getAnnotation(ParamsNotNull.class);
        if (annotation == null) {
            return;
        }
        for (CtParameter<?> param : m.getParameters()) {
            if (param.getType().isPrimitive()) {
                continue;
            }
            if (param.getAnnotation(Default.class) != null) {
                AnnotationHelper.removeAnnotation(param, Default.class);
                continue;
            }
            if (param.getAnnotation(Nullable.class) != null) {
                AnnotationHelper.removeAnnotation(param, Nullable.class);
                continue;
            }
            CtBlock<?> checkBlock = Substitution
                    .substituteMethodBody((CtClass<?>) m.getDeclaringType(), new NullParamTemplate(param, null), "check", getFactory().Type().createReference(Object.class));
            m.getBody().insertBegin(checkBlock);
        }

    }

    protected void removeAnnotations(CtExecutable<?> m) {
        for (CtParameter<?> param : m.getParameters()) {
            AnnotationHelper.removeAnnotation(param, Default.class);
            AnnotationHelper.removeAnnotation(param, Nullable.class);
        }
        AnnotationHelper.removeAnnotation(m, ParamsNotNull.class);
    }

}
